!DEC$ DEFINE rflag=8
subroutine initdir(m,ms,xmax,ymax,zmax,masque,mx,my,mz,matmax)
	integer :: xmax,ymax,zmax,i,j,k,matmax
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real :: mx,my,mz,ms(1:matmax)
    
	m=0
	FORALL(i=0:xmax,j=0:ymax,k=0:zmax,masque(i,j,k).NE.0) 
		m(i,j,k,1)=mx
		m(i,j,k,2)=my
		m(i,j,k,3)=mz
	end forall
	call normvectors(m,ms,masque,xmax,ymax,zmax,matmax)
end subroutine initdir

subroutine initrandom(m,ms,xmax,ymax,zmax,masque,matmax)
	integer :: xmax,ymax,zmax,matmax,i,j,k
	real :: a1,b1,c1,t1,ms(1:matmax)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
	
	CALL RANDOM_SEED()
	m=0

do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
		CALL RANDOM_NUMBER(a1)
		CALL RANDOM_NUMBER(c1)
		CALL RANDOM_NUMBER(b1)
		a1=a1-0.5
		c1=c1-0.5
		b1=b1-0.5
		t1=sqrt((a1*a1)+(c1*c1)+(b1*b1))
		a1=a1/t1
		c1=c1/t1
		b1=b1/t1
		m(i,j,k,1)=a1*ms(masque(i,j,k))
		m(i,j,k,2)=b1*ms(masque(i,j,k))
		m(i,j,k,3)=c1*ms(masque(i,j,k))
end if
end do
end do
end do


end subroutine initrandom

subroutine initvortex(m,ms,xmax,ymax,zmax,masque,cx,cy,matmax)
integer :: xmax,ymax,zmax,matmax,i,j,k
integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
real, INTENT(OUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
real :: r,ms(1:matmax),cx,cy

m=0

do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
		r=sqrt(real((i-cx)*(i-cx)+(j-cy)*(j-cy)))
		if (r>0.9) then
			m(i,j,k,1)=(j-cy)*ms(masque(i,j,k))/r
			m(i,j,k,2)=-(i-cx)*ms(masque(i,j,k))/r
		else
			m(i,j,k,3)=ms(masque(i,j,k))
		end if
end if
end do
end do
end do

end subroutine initvortex

subroutine initvortex9(m,ms,xmax,ymax,zmax,masque,cx,cy,matmax)
integer :: xmax,ymax,zmax,matmax,i,j,k
integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
real, INTENT(OUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
real :: r,ms(1:matmax),cx,cy

m=0

k=0
do i=0,xmax,1
do j=0,ymax,1
if (masque(i,j,k).NE.0) then
		r=sqrt(real((i-cx)*(i-cx)+(j-cy)*(j-cy)))
		if (r>0.9) then
			m(i,j,k,1)=(j-cy)*ms(masque(i,j,k))/r
			m(i,j,k,2)=-(i-cx)*ms(masque(i,j,k))/r
		else
			m(i,j,k,3)=ms(masque(i,j,k))
		end if
end if
end do
end do
k=1
do i=0,xmax,1
do j=0,ymax,1
if (masque(i,j,k).NE.0) then
		r=sqrt(real((i-cx)*(i-cx)+(j-cy)*(j-cy)))
		if (r>0.9) then
			m(i,j,k,1)=-(j-cy)*ms(masque(i,j,k))/r
			m(i,j,k,2)=(i-cx)*ms(masque(i,j,k))/r
		else
			m(i,j,k,3)=ms(masque(i,j,k))
		end if
end if
end do
end do

end subroutine initvortex9


! now for calculating fields and so on

subroutine calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
	integer :: xmax,ymax,zmax,matmax,x,y,z
	real :: ms(matmax),mp(3),beta
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	real, INTENT(INOUT) :: h(0:xmax,0:ymax,0:zmax,1:3)
	real :: temp(0:xmax,0:ymax,0:zmax,1:3), temp2(0:xmax,0:ymax,0:zmax,1:3)

	real :: t1,t2,t3
	integer :: i,j,k,i2,j2,k2,i3,j3

	do i=0,xmax,1
	    do j=0,ymax,1
		do k=0,zmax,1
			if(masque(i,j,k).NE.0) then
		        do i3=1,3,1
			    do i2=0,xmax,1
			    do j2=0,ymax,1
			    do k2=0,zmax,1
			    if (masque(i2,j2,k2).NE.0) then
			    do j3=1,3,1
				h(i,j,k,i3)=h(i,j,k,i3)+aij(i-i2,j-j2,k-k2,i3,j3,masque(i2,j2,k2))*m(i2,j2,k2,j3)
	    		    end do	
			    end if
			    end do
			    end do
			    end do
		    end do
		    end if
		end do
	    end do
	end do
	
	temp2=0
	temp=beta*h
	do i=1,3
	    temp(:,:,:,i)=temp(:,:,:,i)-mp(i)
	end do
	call vectormult(temp2,m,temp,masque,xmax,ymax,zmax)
	do x=0,xmax
	do y=0,ymax
	do z=0,zmax
	    if (masque(x,y,z).ne.0) then
		h(x,y,z,:)=h(x,y,z,:)+temp2(x,y,z,:)/ms(masque(x,y,z))
	    end if
	end do
	end do
	end do
	
end subroutine calcfields


subroutine alignfields(m,masque,h,hold,xmax,ymax,zmax,koeff,ms,matmax)
	integer :: i,j,k
	integer :: xmax,ymax,zmax,matmax
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(IN) :: h(0:xmax,0:ymax,0:zmax,1:3),hold(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(IN) :: ms(1:matmax)
	real :: koeff

	FORALL(i=0:xmax,j=0:ymax,k=0:zmax,masque(i,j,k).NE.0) 
		m(i,j,k,1:3)=h(i,j,k,1:3)*koeff+hold(i,j,k,1:3)*(1-koeff)
	end forall

	call normvectors(m,ms,masque,xmax,ymax,zmax,matmax)

end subroutine alignfields

subroutine normvectors(m,ms,masque,xmax,ymax,zmax,matmax)
	integer :: xmax,ymax,zmax,matmax
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(IN) :: ms(1:matmax)
	integer :: i,j,k
	real :: t1

    do i=0,xmax,1
	do j=0,ymax,1
	    do k=0,zmax,1
		if(masque(i,j,k).NE.0) then
    		    t1=sqrt(m(i,j,k,1)*m(i,j,k,1)+m(i,j,k,2)*m(i,j,k,2)+m(i,j,k,3)*m(i,j,k,3))
    		    m(i,j,k,1:3)=m(i,j,k,1:3)*ms(masque(i,j,k))/t1		    
		end if
	    end do
	end do
    end do
end subroutine normvectors

subroutine getenergy(m,masque,h,xmax,ymax,zmax,energy)
	integer :: xmax,ymax,zmax
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(IN) :: h(0:xmax,0:ymax,0:zmax,1:3)
	integer :: i,j,k,ii,cc
	real :: energy

	energy=0
	cc=0
	do i=0,xmax,1
	do j=0,ymax,1
	    do k=0,zmax,1
		if(masque(i,j,k).NE.0) then
		cc=cc+1
		do ii=1,3,1
		    energy=energy-m(i,j,k,ii)*h(i,j,k,ii)
		end do
		end if
	    end do
	end do
	end do
	energy=energy/cc

end subroutine getenergy

function calcmagn(m,masque,xmax,ymax,zmax,dir)
	integer :: xmax,ymax,zmax,i,j,k
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3)
	integer :: dir
	real :: calcmagn

	calcmagn=0
	do i=0,xmax,1
	    do j=0,ymax,1
		do k=0,zmax,1
		    if(masque(i,j,k).NE.0) then
			calcmagn=calcmagn+m(i,j,k,dir)
    		    end if
    	    end do
	end do
    end do

end function calcmagn

subroutine calcheff(heff,m,h,masque,xmax,ymax,zmax,ms,matmax,beta,mp)
    integer :: xmax,ymax,zmax,matmax
    real :: beta
    real :: mp(3)
    real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3),h(0:xmax,0:ymax,0:zmax,1:3)
    integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
    real :: temp(0:xmax,0:ymax,0:zmax,1:3)
    real, INTENT(INOUT) :: heff(0:xmax,0:ymax,0:zmax,1:3)
    real :: ms(matmax)
    integer :: i,x,y,z

    heff=0
    temp=beta*h
    do i=1,3
	temp(:,:,:,i)=temp(:,:,:,i)-mp(i)
    end do
    call vectormult(heff,m,temp,masque,xmax,ymax,zmax)
    do x=0,xmax
    do y=0,ymax
    do z=0,zmax
	if (masque(x,y,z).ne.0) then
	    heff(x,y,z,:)=h(x,y,z,:)+heff(x,y,z,:)/ms(masque(x,y,z))
	end if
    end do
    end do
    end do
    
    
end subroutine calcheff

subroutine vectormult2(res,vec1,vec2,ms,matmax,masque,xmax,ymax,zmax)
    integer :: xmax,ymax,zmax,matmax
    integer :: i,j,k,i2,j2,k2
    real, INTENT(IN) :: vec1(0:xmax,0:ymax,0:zmax,1:3),vec2(1:3)
    integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
    complex, INTENT(INOUT) :: res(0:xmax,0:ymax,0:zmax,1:3)
    integer :: epsilon(1:3,1:3,1:3)
    real :: coeff,ms(matmax)

    epsilon=0
    epsilon(1,2,3)=1
    epsilon(3,1,2)=1
    epsilon(2,3,1)=1

    epsilon(1,3,2)=-1
    epsilon(3,2,1)=-1
    epsilon(2,1,3)=-1

	do i=0,xmax,1
	do j=0,ymax,1
	do k=0,zmax,1
	if(masque(i,j,k).NE.0) then
	    do i2=1,3,1
	    do j2=1,3,1
		do k2=1,3,1
		    res(i,j,k,i2)=res(i,j,k,i2)+epsilon(i2,j2,k2)*vec1(i,j,k,j2)*vec2(k2)
		end do
	    end do
	    end do
	    res(i,j,k,1:3)=res(i,j,k,1:3)/ms(masque(i,j,k))
	end if
	end do
	end do
	end do	
	
end subroutine vectormult2

subroutine vectormult(res,vec1,vec2,masque,xmax,ymax,zmax)
	integer :: xmax,ymax,zmax
	integer :: i,j,k,i2,j2,k2
	real, INTENT(IN) :: vec1(0:xmax,0:ymax,0:zmax,1:3),vec2(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: res(0:xmax,0:ymax,0:zmax,1:3)
	integer :: epsilon(1:3,1:3,1:3)
	
	epsilon=0

	epsilon(1,2,3)=1
	epsilon(3,1,2)=1
	epsilon(2,3,1)=1

	epsilon(1,3,2)=-1
	epsilon(3,2,1)=-1
	epsilon(2,1,3)=-1
		
	do i=0,xmax,1
	do j=0,ymax,1
	do k=0,zmax,1
	if(masque(i,j,k).NE.0) then
	do i2=1,3,1
	do j2=1,3,1
	    do k2=1,3,1
!		FORALL(i=0:xmax,j=0:ymax,k=0:zmax,i2=1:3,masque(i,j,k).NE.0) 
		    res(i,j,k,i2)=res(i,j,k,i2)+epsilon(i2,j2,k2)*vec1(i,j,k,j2)*vec2(i,j,k,k2)
!		end forall
	    end do
	end do
	end do
	end if
	end do
	end do
	end do	

end subroutine vectormult

! runge-kutta - general functions

subroutine calctorque(m,h,tor,masque,xmax,ymax,zmax)
	integer :: xmax,ymax,zmax,x,y,z
	real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3),h(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: tor(0:xmax,0:ymax,0:zmax,1:3)
!	real :: temptor(0:xmax,0:ymax,0:zmax,1:3)
	
!-mxh
	tor=0
	call vectormult(tor,m,h,masque,xmax,ymax,zmax)

! now -beta*mxmxh
!	temptor=0	
!	call vectormult(temptor,m,tor,masque,xmax,ymax,zmax)
	
!	do x=0,xmax,1
!	do y=0,ymax,1
!	do z=0,zmax,1
!    	    temptor(x,y,z,1:3)=temptor(x,y,z,1:3)/ms(masque(x,y,z))
!	end do
!	end do
!	end do
	
!	temptor=beta*temptor	
	tor=-tor
!-temptor

end subroutine calctorque

subroutine calcdiff(vec1,vec2,masque,xmax,ymax,zmax,error,ms,matmax)
	integer :: xmax,ymax,zmax,matmax
	integer :: i,j,k,j2,cc
	real, INTENT(IN) :: vec1(0:xmax,0:ymax,0:zmax,1:3),vec2(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: error,ms(matmax),t1
	
	error=0
	cc=0
	do i=0,xmax,1
	    do j=0,ymax,1
		do k=0,zmax,1
		    if(masque(i,j,k).NE.0) then
			cc=cc+1
			t1=0
			do j2=1,3,1
!		    	    error=error+abs(vec1(i,j,k,j2)-vec2(i,j,k,j2))
			    t1=t1+abs(vec1(i,j,k,j2)-vec2(i,j,k,j2))
!			    t1=t1+(vec1(i,j,k,j2)-vec2(i,j,k,j2))*(vec1(i,j,k,j2)-vec2(i,j,k,j2))
			end do
!			error=error/ms(masque(i,j,k))
			error=error+t1/ms(masque(i,j,k))
!			error=sqrt(t1)/ms(masque(i,j,k))
    		    end if
    	        end do
	    end do
        end do
	error=error/cc
end subroutine calcdiff

subroutine relax(m,hinit,ms,mp,beta,aij,masque,xmax,ymax,zmax,matmax,koeff,maxerror,tp,aij_fft)
	integer :: xmax,ymax,zmax,matmax,count,tp
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax),ms(1:matmax)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(INOUT) :: m(0:xmax,0:ymax,0:zmax,1:3)
	real :: hold(0:xmax,0:ymax,0:zmax,1:3)
	real, INTENT(INOUT) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
	real :: h(0:xmax,0:ymax,0:zmax,1:3),energy,mp(3),beta
	real :: maxerror,m2(0:xmax,0:ymax,0:zmax,1:3),error,koeff

	error=1D10
	count=0
	hold=hinit
	if (tp.eq.1) then
	    call calcfields(m,ms,mp,beta,masque,aij,hold,xmax,ymax,zmax,matmax)
	else
	    call calcfieldsfft(aij_fft,m,ms,mp,beta,hold,masque,xmax,ymax,zmax,matmax)
	end if
	DO WHILE (error>maxerror)
		m2=m
		h=hinit	
		if (tp.eq.1) then
			call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
		else
			call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
		end if
		call alignfields(m,masque,h,hold,xmax,ymax,zmax,koeff,ms,matmax)
		call calcdiff(m,m2,masque,xmax,ymax,zmax,error,ms,matmax)
		call getenergy(m,masque,h,xmax,ymax,zmax,energy)
		count=count+1
		write(*,*) count,error,energy
		hold=h
	END DO
	hinit=h
end subroutine relax

subroutine get_max(h,masque,xmax,ymax,zmax,max)
	integer :: xmax,ymax,zmax,i,j,k
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real, INTENT(IN) :: h(0:xmax,0:ymax,0:zmax,1:3)
	real :: max,t1,t2

	max=0

	do i=0,xmax,1
		do j=0,ymax,1
		do k=0,zmax,1
			if (masque(i,j,k).NE.0) then
				t1=sqrt(h(i,j,k,1)*h(i,j,k,1)+h(i,j,k,2)*h(i,j,k,2)+h(i,j,k,3)*h(i,j,k,3))
				if (t1>max) max=t1
			end if
		end do
		end do
	end do
		
end subroutine get_max

subroutine writevrml(m,ms,masque,xmax,ymax,zmax,tkoeff,tprint,fname,matmax)
	integer::i,j,k,tprint,matmax,ii
	integer :: xmax,ymax,zmax
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: tkoeff,joff,t1,t2
	real, INTENT(IN) :: m(0:xmax,0:ymax,0:zmax,1:3),ms(1:matmax)
	character(255) :: fname
	integer :: col(1:8,1:3)

	OPEN(UNIT=999,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')
	
	write(999,'(a15)') "#VRML V2.0 utf8"
	write(999,*)
	write(999,*) " PROTO Vector ["
	write(999,*) "field SFVec3f Position 0 0 0"
	write(999,*) "field SFVec3f Direction 0 0 0"
	write(999,*) "field SFFloat StemRadius 0.04"
	write(999,*) "field SFFloat HeadRadius 0.2"

	write(999,*) "field SFFloat HeadSize 0.5"
	write(999,*) "field SFColor Color 1 1 1"
	write(999,*) "]"
	write(999,*) "{"
  	write(999,*) "DEF BasePoint Transform {"
   	write(999,*) "translation IS Position"
   	write(999,*) "children ["

	write(999,*) "DEF Stem Transform { # stem"
    	write(999,*) "children["
	write(999,*) "Shape{"
	write(999,*) "geometry Cylinder {	height 0.5"
	write(999,*) "radius IS StemRadius "
	write(999,*) "top FALSE bottom FALSE}"
	write(999,*) "appearance Appearance {"
	write(999,*) "material Material {diffuseColor IS Color }}"
	write(999,*) "}"
	write(999,*) "]"
	write(999,*) "}"
	
	write(999,*) "DEF Head Transform { # arrowhead"
    	write(999,*) "children["
	write(999,*) "Shape{"
	write(999,*) "geometry Cone	 {	height IS HeadSize"
	write(999,*) "bottomRadius IS HeadRadius }"
	write(999,*) "appearance Appearance {"
	write(999,*) "material Material {diffuseColor 1 0 0 }}"
	write(999,*) "}"
	write(999,*) "	]"
	write(999,*) "}"
   	write(999,*) "]"
	write(999,*) "  }"
	write(999,*) "  DEF Build Script {"
	write(999,*) "	field  SFVec3f	set_direction IS Direction"
	write(999,*) "	field  SFFloat	set_headsize IS HeadSize"
	write(999,*) "	eventOut SFRotation rotation_changed"
	write(999,*) "	eventOut SFVec3f stemcenter_changed"
	write(999,*) "	eventOut SFVec3f stemscale_changed"
	write(999,*) "	eventOut SFVec3f headcenter_changed"
	write(999,*) "url ""vrmlscript:"
	write(999,*) "	function initialize()"
	write(999,*) "{"
	write(999,*) "		phi=0.0;cosphi=0.0;sinphi=0.0;"
	write(999,*) "		nx=0.0;ny=0.0;nz=0.0;"
        write(999,*) 
	write(999,*) "		magnitude=Math.sqrt("
	write(999,*) "			  set_direction[0]*set_direction[0]+"
	write(999,*) "			  set_direction[1]*set_direction[1]+"
	write(999,*) "			  set_direction[2]*set_direction[2]);"
	write(999,*) "		if (magnitude > 0) {"
	write(999,*) "			cosphi = set_direction[1] / magnitude;"
	write(999,*) "		        nx = set_direction[2] / magnitude;"
	write(999,*) "		        ny = 0.0 / magnitude;"
	write(999,*) "			nz = -set_direction[0] / magnitude;"
	write(999,*) "		        sinphi = Math.sqrt(nx * nx + nz * nz);"
	write(999,*) "			}"
	write(999,*) "		if (cosphi != 0) { phi = Math.atan(sinphi / cosphi);}"
 	write(999,*) "			else { phi = Math.PI / 2.0;}"
        write(999,*) "			if (cosphi < 0) { phi += Math.PI;}"
	write(999,*) "			if (sinphi == 0) { nx = 1; } "
	write(999,*) "		else {nx /= sinphi; nz /= sinphi;}"
	write(999,*) "	rotation_changed[0]=nx;"
	write(999,*) "		rotation_changed[1]=ny;"
	write(999,*) "		rotation_changed[2]=nz;"
	write(999,*) "		rotation_changed[3]=phi;"
	write(999,*) "		stemcenter_changed[0]=0;"
	write(999,*) "		stemcenter_changed[1]=(magnitude-set_headsize)/2;"
	write(999,*) "		stemcenter_changed[2]=0;"
	write(999,*) "		stemscale_changed[0]=1.0;"
	write(999,*) "		stemscale_changed[1]=2.0*(magnitude-set_headsize);"
	write(999,*) "		stemscale_changed[2]=1.0;"
	write(999,*) "		headcenter_changed[0]=0;"
	write(999,*) "		headcenter_changed[1]=magnitude-set_headsize/2;"
	write(999,*) "		headcenter_changed[2]=0;"
	write(999,*) "	}"
	write(999,*) "	"" "
 	write(999,*) "}"
	write(999,*) "ROUTE Build.rotation_changed TO BasePoint.rotation"
	write(999,*) "ROUTE Build.stemcenter_changed TO Stem.translation"
	write(999,*) "ROUTE Build.stemscale_changed TO Stem.scale"
	write(999,*) "ROUTE Build.headcenter_changed TO Head.translation"
	write(999,*) "}"

	write(999,*) "Background {"
	write(999,*) "skyColor [1 1 1] groundColor [1 1 1]"
	write(999,*) "}"
	write(999,*) "DEF norm Viewpoint {"
	t1=0
	if (tprint.eq.1) then
	    if (xmax>t1) t1=xmax*tkoeff
	    if (ymax>t1) t1=ymax*tkoeff
	    if (zmax>t1) t1=zmax*tkoeff
	    write(999,'(a8,3F15.2)') "position",xmax*tkoeff/real(2),ymax*tkoeff/real(2),3*t1
	    t2=0
	    if (xmax>t2) t2=xmax*tkoeff
	    if (ymax>t2) t2=ymax*tkoeff
	end if
	if (tprint.eq.2) then
	    if (xmax>t1) t1=xmax*tkoeff
	    if ((ymax+ymax*zmax)>t1) t1=(ymax+ymax*zmax)*tkoeff
	    if (zmax>t1) t1=zmax*tkoeff
	    write(999,'(a8,3F15.2)') "position",xmax*tkoeff/real(2),(ymax+ymax*zmax)*tkoeff/real(2),3*t1
	    t2=0
	    if (xmax>t2) t2=xmax*tkoeff
	    if ((ymax+ymax*zmax)>t2) t2=(ymax+ymax*zmax)*tkoeff
	end if

	    write(999,'(a12,F10.7)') "fieldOfView ",1.2*atan(t2/(2*t1))	

	write(999,*) "}"	


!  the header is complete
	ii=0
	do i=0,1,1
	do j=0,1,1
	do k=0,1,1
	    ii=ii+1
	    col(ii,1)=i
	    col(ii,2)=j
	    col(ii,3)=k	
	end do
	end do
	end do
    if (tprint.eq.1) then
	do i=0,xmax,1
    	do j=0,ymax,1
	do k=0,zmax,1
    	    if (masque(i,j,k).NE.0) then
!		write(*,*) i,j,k,m(i,j,k,1:3)
		if (masque(i,j,k)<=7) then
		    ii=masque(i,j,k)
		else
		    ii=1
		end if
		write(999,'(a18,3F8.2,a12,3F9.5,a7,3I3,a2)') "Vector { Position ", tkoeff*real(i),tkoeff*real(j),tkoeff*real(k)," Direction ", m(i,j,k,1)/ms(masque(i,j,k)),m(i,j,k,2)/ms(masque(i,j,k)),m(i,j,k,3)/ms(masque(i,j,k)), " Color ",col(ii,1:3)," }"
	    end if
	end do
	end do
	end do
	end if

	if (tprint.eq.2) then
	do i=0,xmax,1
	do j=0,ymax,1
	do k=0,zmax,1
	    joff=k*tkoeff*ymax+10
	    if (masque(i,j,k).NE.0) then
		if (masque(i,j,k)<=7) then
		    ii=masque(i,j,k)
		else
		    ii=1
		end if
		write(999,'(a18,3F8.2,a12,3F8.4,a7,3I3,a2)') "Vector { Position ", tkoeff*i,tkoeff*j+joff,tkoeff*k," Direction ", m(i,j,k,1)/ms(masque(i,j,k)),m(i,j,k,2)/ms(masque(i,j,k)),m(i,j,k,3)/ms(masque(i,j,k))," Color ",col(ii,1:3)," }"
	    end if
	end do
	end do
	end do
	end if


        CLOSE(999)
	end subroutine writevrml


subroutine savemt(m,masque,xmax,ymax,zmax,fname)
integer :: xmax,ymax,zmax,fnumber,i,j,k,nn
integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
real, intent(in) :: m(0:xmax,0:ymax,0:zmax,1:3)
character(255) :: fname

OPEN(UNIT=999,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')

nn=0
do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
    nn=nn+1
end if
end do
end do
end do


write(999,'(I9)') nn
	
do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
	write(999,'(3I6,3F20.5)') i,j,k,m(i,j,k,1),m(i,j,k,2),m(i,j,k,3)
end if
end do
end do
end do

CLOSE(999)

end subroutine savemt

subroutine loadmt(m,masque,xmax,ymax,zmax,fname)
integer :: xmax,ymax,zmax,fnumber,i,j,k,nn,ii
integer, INTENT(INOUT) :: masque(0:xmax,0:ymax,0:zmax)
real, intent(out) :: m(0:xmax,0:ymax,0:zmax,1:3)
character(255) :: fname

m=0
OPEN(UNIT=999,FILE=fname,STATUS='OLD',ACTION='READ',FORM='FORMATTED')

read(999,'(I9)') nn
	
do ii=1,nn,1
	read(999,'(3I6,3F20.5)') i,j,k,m(i,j,k,1:3)
end do

CLOSE(999)

end subroutine loadmt

subroutine loadhrft(m,masque,xmax,ymax,zmax,fname)
integer :: xmax,ymax,zmax,fnumber,i,j,k,nn,ii
integer, INTENT(INOUT) :: masque(0:xmax,0:ymax,0:zmax)
complex, intent(out) :: m(0:xmax,0:ymax,0:zmax,1:3)
complex, parameter :: im=(0,1)
real :: t1(1:6)
character(255) :: fname

m=0
OPEN(UNIT=999,FILE=fname,STATUS='OLD',ACTION='READ',FORM='FORMATTED')

read(999,'(I9)') nn
	
do ii=1,nn,1
	read(999,'(3I6,3F20.5)') i,j,k,t1(1:6)
	m(i,j,k,1)=t1(1)+im*t1(2)
	m(i,j,k,2)=t1(3)+im*t1(4)
	m(i,j,k,3)=t1(5)+im*t1(6)
end do

CLOSE(999)

end subroutine loadhrft

subroutine savem(m,masque,xmax,ymax,zmax,fname)
integer :: xmax,ymax,zmax,fnumber,i,j,k,ii,nn
integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
real, intent(in) :: m(0:xmax,0:ymax,0:zmax,1:3)
character(255) :: fname

inquire(iolength=i) ii,j,k,m(ii,j,k,1:3)


OPEN(UNIT=999,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='UNFORMATTED',ACCESS='DIRECT',RECL=i)
nn=0
do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
    nn=nn+1
end if
end do
end do
end do

write(999,REC=1) nn
ii=2

do i=0,xmax,1
do j=0,ymax,1
do k=0,zmax,1
if (masque(i,j,k).NE.0) then
	write(999,REC=ii) i,j,k,m(i,j,k,1:3)
	ii=ii+1
end if
end do
end do
end do

CLOSE(999)

end subroutine savem

subroutine loadm(m,masque,xmax,ymax,zmax,fname)
integer :: xmax,ymax,zmax,fnumber,i,j,k,ii,nn
integer, INTENT(INOUT) :: masque(0:xmax,0:ymax,0:zmax)
real, intent(out) :: m(0:xmax,0:ymax,0:zmax,1:3)
real :: t(3)
character(255) :: fname
m=0
inquire(iolength=i) ii,j,k,m(ii,j,k,1:3)

OPEN(UNIT=999,FILE=fname,STATUS='OLD',ACTION='READ',FORM='UNFORMATTED',ACCESS='DIRECT',RECL=i)

read(999,REC=1) nn

do ii=2,nn+1,1
	read(999,REC=ii) i,j,k,t(1:3)
	if ((i<=xmax).and.(j<=ymax).and.(k<=zmax)) m(i,j,k,1:3)=t(1:3)
end do

CLOSE(999)

end subroutine loadm

subroutine mfft(m,m_fft,masque,xmax,ymax,zmax,matmax)
	use MKL_DFTI
	
	type(DFTI_DESCRIPTOR), POINTER :: Mydescr
	integer :: L(1:3),status,matmax
	integer :: i,j,k,xmax,ymax,zmax,x,y,z
	integer, intent(in) :: masque(0:xmax,0:ymax,0:zmax)
	real, intent(in) :: m(0:xmax,0:ymax,0:zmax,1:3)
	complex, intent(out) :: m_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:matmax)
	complex :: mtemp(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2))
	complex :: X_IN((2*xmax+2)*(2*ymax+2)*(2*zmax+2))
	integer :: strides(4)
	
	L(1)=2*xmax+2
	L(2)=2*ymax+2
	L(3)=2*zmax+2
	
	strides(1)=0
	strides(2)=1
	strides(3)=(2*xmax+2)
	strides(4)=(2*xmax+2)*(2*ymax+2)

	do k=1,matmax,1
	do i=1,3,1
		mtemp=0
		forall(x=1:(xmax+1),y=1:(ymax+1),z=1:(zmax+1),masque(x-1,y-1,z-1).eq.k)
		    mtemp(x,y,z)=m(x-1,y-1,z-1,i)
		end forall

		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
		do z=1,(2*zmax+2),1
    		    X_IN(x+(2*xmax+2)*(y-1)+(2*xmax+2)*(2*ymax+2)*(z-1))=mtemp(x,y,z)
		end do
		end do
		end do
!DEC$ IF (rflag.eq.4)
		status=DftiCreateDescriptor(Mydescr,DFTI_SINGLE,DFTI_COMPLEX,3,L)
!DEC$ ELSE
		status=DftiCreateDescriptor(Mydescr,DFTI_DOUBLE,DFTI_COMPLEX,3,L)
!DEC$ ENDIF
		status=DftiSetValue(Mydescr,DFTI_INPUT_STRIDES,strides)
		status=DftiCommitDescriptor(Mydescr)
		status=DftiComputeForward(Mydescr,X_IN)
		status=DftiFreeDescriptor(Mydescr)
		
		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
		do z=1,(2*zmax+2),1
    		    mtemp(x,y,z)=X_IN(x+(2*xmax+2)*(y-1)+(2*xmax+2)*(2*ymax+2)*(z-1))
		end do
		end do
		end do

!		do x=1,(2*xmax+2),1
!		do y=1,(2*ymax+2),1
!		do z=1,(2*zmax+2),1
!    		    write(*,*) x,y,z,i,k,mtemp(x,y,z)
!		end do
!		end do
!		end do

		m_fft(:,:,:,i,k)=mtemp(:,:,:)
	end do
	end do

end subroutine mfft

subroutine mfft2d(m,m_fft,masque,xmax,ymax,zmax,matmax)
	use MKL_DFTI
	
	type(DFTI_DESCRIPTOR), POINTER :: Mydescr
	integer :: L(1:2),status,matmax
	integer :: i,j,k,xmax,ymax,zmax,x,y,z
	integer, intent(in) :: masque(0:xmax,0:ymax,0:zmax)
	real, intent(in) :: m(0:xmax,0:ymax,0:zmax,1:3)
	complex, intent(out) :: m_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:matmax)
	complex :: mtemp(1:(2*xmax+2),1:(2*ymax+2))
	complex :: X_IN((2*xmax+2)*(2*ymax+2))
	integer :: strides(3)
	
	L(1)=2*xmax+2
	L(2)=2*ymax+2
	
	strides(1)=0
	strides(2)=1
	strides(3)=(2*xmax+2)

	do k=1,matmax,1
	do i=1,3,1
		mtemp=0
		forall(x=1:(xmax+1),y=1:(ymax+1),masque(x-1,y-1,0).eq.k)
		    mtemp(x,y)=m(x-1,y-1,0,i)
		end forall

		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
    		    X_IN(x+(2*xmax+2)*(y-1))=mtemp(x,y)
		end do
		end do
!DEC$ IF (rflag.eq.4)
		status=DftiCreateDescriptor(Mydescr,DFTI_SINGLE,DFTI_COMPLEX,2,L)
!DEC$ ELSE
		status=DftiCreateDescriptor(Mydescr,DFTI_DOUBLE,DFTI_COMPLEX,2,L)
!DEC$ ENDIF
		status=DftiSetValue(Mydescr,DFTI_INPUT_STRIDES,strides)
		status=DftiCommitDescriptor(Mydescr)
		status=DftiComputeForward(Mydescr,X_IN)
		status=DftiFreeDescriptor(Mydescr)
		
		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
    		    mtemp(x,y)=X_IN(x+(2*xmax+2)*(y-1))
		end do
		end do

!		do x=1,(2*xmax+2),1
!		do y=1,(2*ymax+2),1
!		do z=1,(2*zmax+2),1
!    		    write(*,*) x,y,z,i,k,mtemp(x,y,z)
!		end do
!		end do
!		end do

		m_fft(:,:,1,i,k)=mtemp(:,:)
	end do
	end do

end subroutine mfft2d

subroutine hfft(h,h_fft,xmax,ymax,zmax)
	use MKL_DFTI
	
	type(DFTI_DESCRIPTOR), POINTER :: Mydescr
	integer :: L(1:3),status
	integer :: i,j,k,xmax,ymax,zmax,x,y,z
	complex, intent(in) :: h_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3)
	real, intent(inout) :: h(0:xmax,0:ymax,0:zmax,1:3)
	complex :: htemp(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2))
	complex :: X_IN((2*xmax+2)*(2*ymax+2)*(2*zmax+2))
	integer :: strides(4)
	
	L(1)=2*xmax+2
	L(2)=2*ymax+2
	L(3)=2*zmax+2
	strides(1)=0
	strides(2)=1
	strides(3)=(2*xmax+2)
	strides(4)=(2*xmax+2)*(2*ymax+2)

	do i=1,3,1
		htemp=0
		htemp(:,:,:)=h_fft(:,:,:,i)
		
		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
		do z=1,(2*zmax+2),1
    		    X_IN(x+(2*xmax+2)*(y-1)+(2*xmax+2)*(2*ymax+2)*(z-1))=htemp(x,y,z)
		end do
		end do
		end do

!DEC$ IF (rflag.eq.4)
		status=DftiCreateDescriptor(Mydescr,DFTI_SINGLE,DFTI_COMPLEX,3,L)
!DEC$ ELSE
		status=DftiCreateDescriptor(Mydescr,DFTI_DOUBLE,DFTI_COMPLEX,3,L)
!DEC$ ENDIF
		status=DftiSetValue(Mydescr,DFTI_INPUT_STRIDES,strides)
		status=DftiSetValue(Mydescr,DFTI_BACKWARD_SCALE,1/real(L(1)*L(2)*L(3)))
		status=DftiCommitDescriptor(Mydescr)
		status=DftiComputeBackward(Mydescr,X_IN)
		status=DftiFreeDescriptor(Mydescr)

		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
		do z=1,(2*zmax+2),1
    		    htemp(x,y,z)=X_IN(x+(2*xmax+2)*(y-1)+(2*xmax+2)*(2*ymax+2)*(z-1))
		end do
		end do
		end do
		
!		h(:,:,:,i)=h(:,:,:)+htemp(:,:,:)		
		h(0:xmax,0:ymax,0:zmax,i)=h(0:xmax,0:ymax,0:zmax,i)+real(htemp(1:(xmax+1),1:(ymax+1),1:(zmax+1)))
	end do

end subroutine hfft

subroutine hfft2d(h,h_fft,xmax,ymax,zmax)
	use MKL_DFTI
	
	type(DFTI_DESCRIPTOR), POINTER :: Mydescr
	integer :: L(1:2),status
	integer :: i,j,k,xmax,ymax,zmax,x,y,z
	complex, intent(in) :: h_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3)
	real, intent(inout) :: h(0:xmax,0:ymax,0:zmax,1:3)
	complex :: htemp(1:(2*xmax+2),1:(2*ymax+2))
	complex :: X_IN((2*xmax+2)*(2*ymax+2))
	integer :: strides(3)
	
	L(1)=2*xmax+2
	L(2)=2*ymax+2
	strides(1)=0
	strides(2)=1
	strides(3)=(2*xmax+2)

	do i=1,3,1
		htemp=0
		htemp(:,:)=h_fft(:,:,1,i)
		
		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
    		    X_IN(x+(2*xmax+2)*(y-1))=htemp(x,y)
		end do
		end do

!DEC$ IF (rflag.eq.4)
		status=DftiCreateDescriptor(Mydescr,DFTI_SINGLE,DFTI_COMPLEX,2,L)
!DEC$ ELSE
		status=DftiCreateDescriptor(Mydescr,DFTI_DOUBLE,DFTI_COMPLEX,2,L)
!DEC$ ENDIF

		status=DftiSetValue(Mydescr,DFTI_INPUT_STRIDES,strides)
		status=DftiSetValue(Mydescr,DFTI_BACKWARD_SCALE,1/real(L(1)*L(2)))
		status=DftiCommitDescriptor(Mydescr)
		status=DftiComputeBackward(Mydescr,X_IN)
		status=DftiFreeDescriptor(Mydescr)

		do x=1,(2*xmax+2),1
		do y=1,(2*ymax+2),1
    		    htemp(x,y)=X_IN(x+(2*xmax+2)*(y-1))
		end do
		end do
		
!		h(:,:,:,i)=h(:,:,:)+htemp(:,:,:)		
		h(0:xmax,0:ymax,0,i)=h(0:xmax,0:ymax,0,i)+real(htemp(1:(xmax+1),1:(ymax+1)))
	end do

end subroutine hfft2d

subroutine calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
	integer :: xmax,ymax,zmax,matmax
	real :: ms(matmax),mp(3),beta
	integer, intent(in) :: masque(0:xmax,0:ymax,0:zmax)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real, intent(inout) :: h(0:xmax,0:ymax,0:zmax,1:3)
	real, intent(in) :: m(0:xmax,0:ymax,0:zmax,1:3)
	complex :: m_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:matmax),h_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3)
	integer :: i,j,x,y,z,k
	real :: temp(0:xmax,0:ymax,0:zmax,1:3), temp2(0:xmax,0:ymax,0:zmax,1:3)

! now this includes 2 version - standart for zmax>0 and spetial 2d for zmax=0
	if (zmax>0) then
	    call mfft(m,m_fft,masque,xmax,ymax,zmax,matmax)

	    h_fft=0
	    do x=1,2*xmax+2,1
	    do y=1,2*ymax+2,1
	    do z=1,2*zmax+2,1
	    do i=1,3,1
	    do j=1,3,1
	    do k=1,matmax,1
		h_fft(x,y,z,i)=h_fft(x,y,z,i)+aij_fft(x,y,z,i,j,k)*m_fft(x,y,z,j,k)
	    end do
	    end do
	    end do
	    end do
	    end do
	    end do

	    call hfft(h,h_fft,xmax,ymax,zmax)
	else
	    call mfft2d(m,m_fft,masque,xmax,ymax,zmax,matmax)

	    h_fft=0
	    do x=1,2*xmax+2,1
	    do y=1,2*ymax+2,1
	    do i=1,3,1
	    do j=1,3,1
	    do k=1,matmax,1
		h_fft(x,y,1,i)=h_fft(x,y,1,i)+aij_fft(x,y,1,i,j,k)*m_fft(x,y,1,j,k)
	    end do
	    end do
	    end do
	    end do
	    end do

	    call hfft2d(h,h_fft,xmax,ymax,zmax)	
	end if
	
	temp2=0
	temp=beta*h
	do i=1,3
	    temp(:,:,:,i)=temp(:,:,:,i)-mp(i)
	end do
	call vectormult(temp2,m,temp,masque,xmax,ymax,zmax)
	do x=0,xmax
	do y=0,ymax
	do z=0,zmax
	    if (masque(x,y,z).ne.0) then
		h(x,y,z,:)=h(x,y,z,:)+temp2(x,y,z,:)/ms(masque(x,y,z))
	    end if
	end do
	end do
	end do
	
end subroutine calcfieldsfft

subroutine print_red_m(m,xmax,ymax,zmax,masque,tkoeff,tprint,scx,scy,scz,fname)
    integer :: xmax,ymax,zmax
    real :: m(0:xmax,0:ymax,0:zmax,1:3),tkoeff,ms2(1:1),mtemp(1:3),t1
    integer :: masque(0:xmax,0:ymax,0:zmax),tprint,scx,scy,scz
    character(255) :: fname
    real, allocatable :: m2(:,:,:,:)
    integer,allocatable :: masque2(:,:,:)
    integer :: x2,y2,z2,x,y,z,xm,ym,zm,n,xx,yy,zz
    
    
    ms2=1
    xm=nint((xmax/real(scx))+0.49)
    ym=nint((ymax/real(scy))+0.49)
    zm=nint((zmax/real(scz))+0.49)
    allocate(masque2(0:xm,0:ym,0:zm),m2(0:xm,0:ym,0:zm,1:3))
    m2=0
    masque2=0

    do x=0,xmax,scx
    do y=0,ymax,scy
    do z=0,zmax,scz
    n=0
    mtemp=0
	do x2=0,scx-1,1
	do y2=0,scy-1,1
	do z2=0,scz-1,1
	    xx=x+x2
	    yy=y+y2
	    zz=z+z2
	    if ((xx<=xmax).and.(yy<=ymax).and.(zz<=zmax)) then
		if (masque(xx,yy,zz).NE.0) then
!		write(*,*) xx,yy,zz,n+1
		    n=n+1
		    mtemp(:)=mtemp(:)+m(xx,yy,zz,:)
		end if
	    end if
	end do
	end do
	end do
	if (n>0) then
	    mtemp=mtemp/n
	    t1=sqrt(mtemp(1)*mtemp(1)+mtemp(2)*mtemp(2)+mtemp(3)*mtemp(3))
	    mtemp=mtemp/t1
	    masque2((x/real(scx)),nint(y/real(scy)),nint(z/real(scz)))=1
	    m2(nint(x/real(scx)),nint(y/real(scy)),nint(z/real(scz)),:)=mtemp(:)
	end if
    end do
    end do
    end do
    
    call writevrml(m2,ms2,masque2,xm,ym,zm,tkoeff,tprint,fname,int(1))
    
end subroutine print_red_m

subroutine inith(h,xmax,ymax,zmax,omega,time,hinit,hrf)
    integer :: xmax,ymax,zmax
    real :: omega,time,mpinit(3),mp(3)
    real, INTENT(INOUT) :: h(0:xmax,0:ymax,0:zmax,1:3)
    real, INTENT(IN) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
    complex, parameter :: im=(0,1)
    complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
    integer :: i

    h=hinit
    h(:,:,:,1:3)=h(:,:,:,1:3)+real(hrf(:,:,:,1:3)*exp(-im*omega*time))
!    write(100,*) h(:,:,:,1:3)
end subroutine inith
    
subroutine rm(m,masque,xmax,ymax,zmax,ms,matmax,pert)
    integer :: xmax,ymax,zmax,matmax,i,j,k,ii
    integer :: masque(0:xmax,0:ymax,0:zmax)
    real :: m(0:xmax,0:ymax,0:zmax,1:3)
    real :: ms(1:matmax)
    real :: a(1:3),pert
    
    CALL RANDOM_SEED()

    do i=0,xmax,1
    do j=0,ymax,1
    do k=0,zmax,1
    if (masque(i,j,k).NE.0) then
	do ii=1,3
	    CALL RANDOM_NUMBER(a(ii))
	end do
	a=a-0.5
	a=a*ms(masque(i,j,k))*pert/sqrt(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
	do ii=1,3
	    m(i,j,k,ii)=m(i,j,k,ii)+a(ii)
	end do
    end if
    end do
    end do
    end do
    call normvectors(m,ms,masque,xmax,ymax,zmax,matmax)
end subroutine rm

subroutine fm(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,gyro,timestep,aij,aij_fft,hinit,hrf,omega,tp,endtime,fname,scvx,scvy,scvz,vrtype,cmax)
! tp=1 - simple
!=2 - fft
! tp=1 - till count<endtime
! tp=2 - till time<endtime
	integer :: xmax,ymax,zmax,matmax,iii,x,y,z,i,tp,tp2,check
	integer :: scvx,scvy,scvz,vrtype,cc
	real :: cmax
	complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
	real :: beta,timestep,error,energy,omega,endtime,gyro,time,mp(1:3),timer,speed,ph2
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real :: m(0:xmax,0:ymax,0:zmax,1:3),timr,ene,en
	real :: ms(1:matmax),count,phase,mpp(3)
	real, INTENT(INOUT) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: h(0:xmax,0:ymax,0:zmax,1:3)
	real :: ctime,ctime2
	character(255) :: fname,fname2

	OPEN(UNIT=990,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')
	!!!!
!	call pmode(m,masque,xmax,ymax,zmax,ms,matmax,real(0.1))
	!!!!!
	time=0
	check=0
	count=0
	cc=0
	call CPU_TIME(ctime)
	DO WHILE (check.eq.0)
		call calcp(m,mpp,masque,xmax,ymax,zmax,ms(1))
		write(990,'(F25.14,3F12.6)',ADVANCE='NO') time/gyro,mpp(1:3)
		call iterate(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,timestep,aij,aij_fft,hinit,hrf,omega,tp,time,energy)
		write(990,'(F25.3)') energy
		if ((time/gyro)>endtime) check=1
		time=time+timestep
		count=count+timestep
		if ((cmax>0).and.(count>=cmax)) then
		cc=cc+1
		    count=0
		    write(fname2,'(I0,a4)') cc,".wrl"
		    call print_red_m(m,xmax,ymax,zmax,masque,real(1),vrtype,scvx,scvy,scvz,fname2)
		end if
	END DO
	call CPU_TIME(ctime2)
	write(*,*) "time spent ",ctime2-ctime
	CLOSE(990)
end subroutine fm

subroutine relaxrk4(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,gyro,timestep,aij,aij_fft,hinit,tp,error)
! tp=1 - simple
!=2 - fft
! tp=1 - till count<endtime
! tp=2 - till time<endtime
	integer :: xmax,ymax,zmax,matmax,iii,x,y,z,i,tp,tp2,check
	integer :: scvx,scvy,scvz,vrtype,cc
	real :: cmax,error
	complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
	real :: beta,timestep,energy,omega,endtime,gyro,time,mp(1:3),timer,speed,ph2
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real :: m(0:xmax,0:ymax,0:zmax,1:3),timr,ene,en
	real :: ms(1:matmax),count,phase,mpp(3)
	real, INTENT(INOUT) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: h(0:xmax,0:ymax,0:zmax,1:3)
	real :: ctime,ctime2,err,enold,emin
	integer :: st
!
	real :: mmin(0:xmax,0:ymax,0:zmax,1:3)

	hrf=0
	omega=0
	time=0
	check=0
	count=0
	cc=0
	st=0
	emin=1D200
	call CPU_TIME(ctime)
	DO WHILE (check.eq.0)
		call iterate(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,timestep,aij,aij_fft,hinit,hrf,omega,tp,time,energy)
		write(*,*) time/gyro,energy
		if (energy<emin) then
		    emin=energy
!		    
		    mmin=m
		end if
		if (time>(timestep/3)) st=1
		err=(enold-energy)/abs(enold)
		time=time+timestep
		enold=energy
		if ((err<error).and.(err>0).and.(st.eq.1)) check=1
	END DO
!
!	m=mmin
	call CPU_TIME(ctime2)
	write(*,*) "time spent ",ctime2-ctime
end subroutine relaxrk4

subroutine set_timestep(hinit,hrf,ms,mp,beta,aij,aij_fft,masque,xmax,ymax,zmax,matmax,timestep,omega)
	integer :: xmax,ymax,zmax,matmax,x,y,z,tp,n
	integer :: scvx,scvy,scvz,vrtype
	real :: cmax,gyro
	real :: timestep,endtime,mt,mtmax,ms(matmax),mp(3),beta,error,omega
	integer :: masque(0:xmax,0:ymax,0:zmax),check
	real :: hinit(0:xmax,0:ymax,0:zmax,1:3),m(0:xmax,0:ymax,0:zmax,1:3),h(0:xmax,0:ymax,0:zmax,1:3)
	real :: m2(0:xmax,0:ymax,0:zmax,1:3),m3(0:xmax,0:ymax,0:zmax,1:3)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
	character(255) :: fname
	
	cmax=-1
	gyro=1
	call initrandom(m,ms,xmax,ymax,zmax,masque,matmax)
	fname="test.dat"
	call inith(h,xmax,ymax,zmax,omega,real(0),hinit,hrf)
	call get_n(masque,xmax,ymax,zmax,n)
	if (n>300) then
	    tp=2
	else
	    tp=1
	end if
	if (tp.eq.2) then
	    call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
	else
	    call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
	end if
	call get_max(h,masque,xmax,ymax,zmax,mtmax)
	timestep=1/(16*mtmax)
	write(*,*) timestep
	endtime=128*timestep
	check=0
	m2=m
	call fm(m2,ms,mp,masque,xmax,ymax,zmax,matmax,beta,gyro,timestep,aij,aij_fft,hinit,hrf,omega,tp,endtime,fname,scvx,scvy,scvz,vrtype,cmax)
	do while(check.eq.0)
	    timestep=2*timestep
	    m3=m
	    call fm(m3,ms,mp,masque,xmax,ymax,zmax,matmax,beta,gyro,timestep,aij,aij_fft,hinit,hrf,omega,tp,endtime,fname,scvx,scvy,scvz,vrtype,cmax)
	    call calcdiff(m3,m2,masque,xmax,ymax,zmax,error,ms,matmax)
	    write(*,*) timestep,error
	    if (error>0.01) check=1
	end do	
	timestep=timestep/2
end subroutine set_timestep

subroutine iterate(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,timestep,aij,aij_fft,hinit,hrf,omega,tp,startime,energy)
! tp=1 - simple
!=2 - fft
! tp=1 - till count<endtime
! tp=2 - till time<endtime
	integer :: xmax,ymax,zmax,matmax,iii,x,y,z,i,tp,tp2,check
	complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
	real :: beta,timestep,error,energy,omega,endtime,time,startime,ph2
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real :: m(0:xmax,0:ymax,0:zmax,1:3)
	real :: ms(1:matmax),count,phase,mpp(3),mp(3)
	real, INTENT(INOUT) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: m0(0:xmax,0:ymax,0:zmax,1:3),tor(0:xmax,0:ymax,0:zmax,1:3),h(0:xmax,0:ymax,0:zmax,1:3),timer
	real :: k1(0:xmax,0:ymax,0:zmax,1:3),k2(0:xmax,0:ymax,0:zmax,1:3),k3(0:xmax,0:ymax,0:zmax,1:3),k4(0:xmax,0:ymax,0:zmax,1:3)

	phase=0
	check=0
	m0=m
	time=startime
	timer=startime
!		write(10,*) m(0,0,0,1:3)	
		call inith(h,xmax,ymax,zmax,omega,time,hinit,hrf)
!		write(10,*) h(0,0,0,1:3)	
		if (tp.eq.1) then
		    call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
		else
		    call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
		end if
!		write(10,*) h(0,0,0,1:3)
		call getenergy(m,masque,h,xmax,ymax,zmax,energy)
		call calctorque(m,h,tor,masque,xmax,ymax,zmax,beta,ms,matmax)
!		write(10,*) tor(0,0,0,1:3)
		k1=timestep*tor
		time=timer+(timestep/2)
		call inith(h,xmax,ymax,zmax,omega,time,hinit,hrf)
		m=m0+k1/2

		if (tp.eq.1) then
		    call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
		else
		    call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
		end if

		call calctorque(m,h,tor,masque,xmax,ymax,zmax,beta,ms,matmax)
		k2=timestep*tor

		time=timer+(timestep/2)
		call inith(h,xmax,ymax,zmax,omega,time,hinit,hrf)
		m=m0+k2/2

		if (tp.eq.1) then
		    call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
		else
		    call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
		end if
		call calctorque(m,h,tor,masque,xmax,ymax,zmax,beta,ms,matmax)
		k3=timestep*tor

		time=timer+timestep
		call inith(h,xmax,ymax,zmax,omega,time,hinit,hrf)
		m=m0+k3

		if (tp.eq.1) then
		    call calcfields(m,ms,mp,beta,masque,aij,h,xmax,ymax,zmax,matmax)
		else
		    call calcfieldsfft(aij_fft,m,ms,mp,beta,h,masque,xmax,ymax,zmax,matmax)
		end if
		call calctorque(m,h,tor,masque,xmax,ymax,zmax,beta,ms,matmax)
		k4=timestep*tor
!		write(10,*) "results"
		m=m0+((k1+k4)/6)+((k2+k3)/3)
!		write(10,*) m(0,0,0,1:3)
		call normvectors(m,ms,masque,xmax,ymax,zmax,matmax)
		
end subroutine iterate

subroutine calcp(m0,mpp,masque,xmax,ymax,zmax,ms)
    integer :: xmax,ymax,zmax,x,y,z
    integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
    real :: m0(0:xmax,0:ymax,0:zmax,1:3),mpp(3),ms
    integer :: i
	i=0
	mpp=0
	do x=0,xmax
    	do y=0,ymax
	do z=0,zmax
	    if (masque(x,y,z)>0) then
    	        mpp(:)=mpp(:)+m0(x,y,z,:)
	        i=i+1
	    end if
	end do
	end do
	end do
	mpp=mpp/(i*ms)
    
end subroutine calcp

subroutine fm3(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,gyro,timestep,aij,aij_fft,hinit,hrf,omega,tp,endtime,fname,modes2,n,nn,num,cmax)
! tp=1 - simple
!=2 - fft
! tp=1 - till count<endtime
! tp=2 - till time<endtime
	integer :: xmax,ymax,zmax,matmax,iii,x,y,z,i,j,tp,tp2,check,n,nit,ii,nn
	real :: cmax
	complex :: hrf(0:xmax,0:ymax,0:zmax,1:3)
	real :: beta,timestep,error,energy,omega,endtime,gyro,time,mp(1:3),timer,speed,ph2
	real, INTENT(IN) :: aij(-xmax:(xmax+1),-ymax:(ymax+1),-zmax:(zmax+1),1:3,1:3,1:matmax)
	complex, intent(in) :: aij_fft(1:(2*xmax+2),1:(2*ymax+2),1:(2*zmax+2),1:3,1:3,1:matmax)
	real :: m(0:xmax,0:ymax,0:zmax,1:3),timr,ene,en,m0(0:xmax,0:ymax,0:zmax,1:3),dm(0:xmax,0:ymax,0:zmax,1:3)
	real :: ms(1:matmax),count,phase,mpp(3)
	real, INTENT(INOUT) :: hinit(0:xmax,0:ymax,0:zmax,1:3)
	integer, INTENT(IN) :: masque(0:xmax,0:ymax,0:zmax)
	real :: h(0:xmax,0:ymax,0:zmax,1:3)
	real :: ctime,ctime2
	real :: maxes(1:nn),gmax,t1
	real, allocatable :: a(:,:),aphase(:,:)
	integer :: num(0:xmax,0:ymax,0:zmax)
	complex :: modes2(1:n,1:3,nn),tt
	character(255) :: fname
	
	time=0
	check=0
	count=cmax*gyro+1
	nit=nint(endtime/cmax)+1
	allocate(a(nit,1:nn),aphase(nit,1:nn))
	a=0
	aphase=0
	call CPU_TIME(ctime)	
	m0=m
!	call rm(m,masque,xmax,ymax,zmax,ms,matmax,real(0.1))
!	call pmode(m,masque,xmax,ymax,zmax,ms,matmax,real(0.1))
	gmax=0
	maxes=0
	OPEN(UNIT=990,FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')
	ii=0
	OPEN(UNIT=993,FILE="modes.dat",STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')	
	OPEN(UNIT=994,FILE="phases.dat",STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')	
!	write(993,'(a15)', ADVANCE='NO') "               "
!	write(994,'(a15)', ADVANCE='NO') "               "
!	do j=1,nn
!	    write(993,'(a7,I0,a7)', ADVANCE='NO') "   mode",j,"       "
!	    write(994,'(a5,I0,a2)', ADVANCE='NO') " mode",j,"  "
!	end do
!	write(993,*)
!	write(994,*)

	DO WHILE (check.eq.0)
	    call calcp(m,mpp,masque,xmax,ymax,zmax,ms(1))
	    write(990,'(F25.14,3F12.6)',ADVANCE='NO') time/gyro,mpp(1:3)
	    call iterate(m,ms,mp,masque,xmax,ymax,zmax,matmax,beta,timestep,aij,aij_fft,hinit,hrf,omega,tp,time,energy)
	    write(990,'(F25.3)') energy
	    if ((time/gyro)>endtime) check=1
	    time=time+timestep
	    count=count+timestep
	    if ((count/gyro)>=cmax) then
	        ii=ii+1
	        count=0
	        dm=m-m0
		write(993,'(F15.10)', ADVANCE='NO') time/gyro
		write(994,'(F15.10)', ADVANCE='NO') time/gyro
		do j=1,nn
		    tt=0
		    do x=0,xmax
		    do y=0,ymax
		    do z=0,zmax
		    if (masque(x,y,z).ne.0) then
		        do iii=1,3
		    	    tt=tt+dm(x,y,z,iii)*modes2(num(x,y,z),iii,j)
			end do
		    end if
		    end do
		    end do
		    end do
		    t1=abs(tt)
		    a(ii,j)=t1
		    if (t1>gmax) gmax=t1
		    if (t1>maxes(j)) maxes(j)=t1
		    call atg(t1,imag(tt),real(tt))
		    aphase(ii,j)=t1
		    write(993,'(F15.3)', ADVANCE='NO') a(ii,j)
		    write(994,'(F9.5)', ADVANCE='NO') aphase(ii,j)
		end do		    
		write(993,*)
		write(994,*)

		end if
	END DO
	call CPU_TIME(ctime2)
	write(*,*) "time spent ",ctime2-ctime
	close(993)
	close(994)

	time=0	
	OPEN(UNIT=991,FILE="modesmax.dat",STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')
	do j=1,nn
	    write(991,'(F15.3)') maxes(j)
	end do
	close(991)
	
	OPEN(UNIT=991,FILE="modes01.dat",STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')
	OPEN(UNIT=992,FILE="modes03.dat",STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')

	write(991,'(a15)', ADVANCE='NO') "               "
	write(992,'(a15)', ADVANCE='NO') "               "

	do j=1,nn
	    if (maxes(j)>(gmax*0.1)) write(991,'(a7,I0,a7)', ADVANCE='NO') "   mode",j,"       "
	    if (maxes(j)>(gmax*0.5)) write(992,'(a7,I0,a7)', ADVANCE='NO') "   mode",j,"       "
	end do
	write(991,*)
	write(992,*)
	do i=1,ii
	    time=(i-1)*cmax*gyro
! 10%
	    write(991,'(F15.10)', ADVANCE='NO') time/gyro
! 50%
	    write(992,'(F15.10)', ADVANCE='NO') time/gyro
! all
	    do j=1,nn
		if (maxes(j)>(gmax*0.5)) write(992,'(F15.3)', ADVANCE='NO') a(i,j)
		if (maxes(j)>(gmax*0.1)) write(991,'(F15.3)', ADVANCE='NO') a(i,j)
	    end do
	    write(991,*)
	    write(992,*)
	end do
	close(990)
	close(991)
	close(992)
	do j=1,nn
	    if(maxes(j)>(gmax*0.25)) then
		write(fname,'(a4,I0,a4)') "mode",j,".dat"
		OPEN(UNIT=(300+j),FILE=fname,STATUS='REPLACE',ACTION='WRITE',FORM='FORMATTED')		
	    end if
	end do
	do j=1,nn
	    if(maxes(j)>(gmax*0.25)) then
	    do i=1,ii
		time=(i-1)*cmax*gyro
		write(300+j,'(F15.10,F15.3,F9.5)') time/gyro,a(i,j),aphase(i,j)
	    end do
	    end if
	end do
	
	do j=1,nn
	    if(maxes(j)>(gmax*0.25)) then
		close(300+j)	
	    end if
	end do	

end subroutine fm3

subroutine pmode(m,masque,xmax,ymax,zmax,ms,matmax,pert)
    integer :: xmax,ymax,zmax,matmax,i,j,k,ii
    integer :: masque(0:xmax,0:ymax,0:zmax)
    real :: m(0:xmax,0:ymax,0:zmax,1:3)
    real :: ms(1:matmax)
    real :: a(1:3),pert
    
    CALL RANDOM_SEED()

    do i=0,xmax,1
    do j=0,ymax,1
    do k=0,zmax,1
    if (masque(i,j,k).NE.0) then
	do ii=2,2
	    m(i,j,k,ii)=m(i,j,k,ii)+pert*ms(masque(i,j,k))
	end do
    end if
    end do
    end do
    end do
    call normvectors(m,ms,masque,xmax,ymax,zmax,matmax)
end subroutine pmode
